#!/usr/bin/python3
# -*- coding: utf-8 -*-

from base64 import b64decode
from re import compile as re_compile
from time import sleep
from traceback import format_exc
from threading import Thread, Lock
from typing import Any, Dict, List, Optional

from API import API  # type: ignore
from Controller import Controller

from docker import DockerClient
from docker.models.services import Service


class SwarmController(Controller):
    def __init__(
        self,
        docker_host: str,
        core_api: API,
        log_level: str = "INFO",
        *,
        api_token: Optional[str] = None,
        wait_retry_interval: int = 5,
    ):
        super().__init__(
            "swarm",
            core_api,
            log_level=log_level,
            api_token=api_token,
            wait_retry_interval=wait_retry_interval,
        )
        self.__client = DockerClient(base_url=docker_host)
        self.__internal_lock = Lock()
        self.__env_rx = re_compile(r"^(?![#\s])([^=]+)=([^\n]*)$")

    def _get_controller_instances(self) -> List[Service]:
        return self.__client.services.list(filters={"label": "bunkerweb.INSTANCE"})  # type: ignore

    def _get_controller_services(self) -> List[Service]:
        return self.__client.services.list(filters={"label": "bunkerweb.SERVER_NAME"})  # type: ignore

    def _to_instances(self, controller_instance) -> List[dict]:
        instances = []
        instance_env = {}
        for env in controller_instance.attrs["Spec"]["TaskTemplate"]["ContainerSpec"]["Env"]:
            match = self.__env_rx.search(env)
            if match is None:
                continue
            groups = match.groups()
            if self._is_setting(groups[0]):
                instance_env[groups[0]] = groups[1]

        for task in controller_instance.tasks():
            if task["DesiredState"] != "running":
                continue
            instances.append(
                {
                    "name": task["ID"],
                    "hostname": f"{controller_instance.name}.{task['NodeID']}.{task['ID']}",
                    "health": task["Status"]["State"] == "running",
                    "env": instance_env,
                }
            )
        return instances

    def _to_services(self, controller_service) -> List[dict]:
        service = {}
        for variable, value in controller_service.attrs["Spec"]["Labels"].items():
            if not variable.startswith("bunkerweb."):
                continue
            real_variable = variable.replace("bunkerweb.", "", 1)
            if not self._is_setting_context(real_variable, "multisite"):
                continue
            service[real_variable] = value
        return [service]

    def _get_static_services(self) -> List[dict]:
        services = []
        variables = {}
        for instance in self.__client.services.list(filters={"label": "bunkerweb.INSTANCE"}):
            if not instance.attrs or not instance.attrs.get("Spec", {}).get("TaskTemplate", {}).get("ContainerSpec", {}).get("Env"):
                continue

            for env in instance.attrs["Spec"]["TaskTemplate"]["ContainerSpec"]["Env"]:
                match = self.__env_rx.search(env)
                if match is None:
                    continue
                groups = match.groups()
                variables[groups[0]] = groups[1]
        if "SERVER_NAME" in variables and variables["SERVER_NAME"].strip():
            for server_name in variables["SERVER_NAME"].strip().split():
                service = {}
                service["SERVER_NAME"] = server_name
                for variable, value in variables.items():
                    prefix = variable.split("_")[0]
                    real_variable = variable.replace(f"{prefix}_", "", 1)
                    if prefix == server_name and self._is_setting_context(real_variable, "multisite"):
                        service[real_variable] = value
                services.append(service)
        return services

    def get_configs(self) -> Dict[str, Dict[str, Any]]:
        configs = {}
        for config_type in self._supported_config_types:
            configs[config_type] = {}
        for config in self.__client.configs.list(filters={"label": "bunkerweb.CONFIG_TYPE"}):
            if not config.name or not config.attrs or not config.attrs.get("Spec", {}).get("Labels", {}) or not config.attrs.get("Spec", {}).get("Data", {}):
                continue

            config_type = config.attrs["Spec"]["Labels"]["bunkerweb.CONFIG_TYPE"]
            config_name = config.name
            if config_type not in self._supported_config_types:
                self._logger.warning(
                    f"Ignoring unsupported CONFIG_TYPE {config_type} for Config {config_name}",
                )
                continue
            config_site = ""
            if "bunkerweb.CONFIG_SITE" in config.attrs["Spec"]["Labels"]:
                if not self._is_service_present(config.attrs["Spec"]["Labels"]["bunkerweb.CONFIG_SITE"]):
                    self._logger.warning(
                        f"Ignoring config {config_name} because {config.attrs['Spec']['Labels']['bunkerweb.CONFIG_SITE']} doesn't exist",
                    )
                    continue
                config_site = f"{config.attrs['Spec']['Labels']['bunkerweb.CONFIG_SITE']}/"
            configs[config_type][f"{config_site}{config_name}"] = b64decode(config.attrs["Spec"]["Data"])
        return configs

    def apply_config(self) -> bool:
        return self.apply(
            self._instances,
            self._services,
            configs=self._configs,
            first=not self._loaded,
        )

    def __event(self, event_type):
        while True:
            locked = False
            error = False
            try:
                for _ in self.__client.events(decode=True, filters={"type": event_type}):
                    self.__internal_lock.acquire()
                    locked = True
                    try:
                        self._instances = self.get_instances()
                        self._services = self.get_services()
                        self._configs = self.get_configs()
                        if not self.update_needed(self._instances, self._services, configs=self._configs):
                            self.__internal_lock.release()
                            locked = False
                            sleep(1)
                            continue
                        self._logger.info(f"Caught Swarm event ({event_type}), deploying new configuration ...")
                        if not self.apply_config():
                            self._logger.error("Error while deploying new configuration")
                        else:
                            self._logger.info(
                                "Successfully deployed new configuration 🚀",
                            )
                    except:
                        self._logger.error(f"Exception while processing Swarm event ({event_type}) :\n{format_exc()}")
                    self.__internal_lock.release()
                    locked = False
            except:
                self._logger.error(
                    f"Exception while reading Swarm event ({event_type}) :\n{format_exc()}",
                )
                error = True
            finally:
                if locked:
                    self.__internal_lock.release()
                    locked = False
                if error is True:
                    self._logger.warning("Got exception, retrying in 10 seconds ...")
                    sleep(10)

    def process_events(self):
        event_types = ("service", "config")
        threads = [Thread(target=self.__event, args=(event_type,)) for event_type in event_types]
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
